from __future__ import annotations

import builtins
import math
import re
from logging import Logger
from typing import List, Optional, Tuple

import pcbnew

from .board_modifier import KICAD_VERSION, BoardModifier
from .element_position import ElementInfo, ElementPosition, Point, PositionOption, Side
from .kle_serial import get_keyboard


def position_in_rotated_coordinates(
    point: pcbnew.wxPoint, angle: float
) -> pcbnew.wxPoint:
    """
    Map position in xy-Cartesian coordinate system to x'y'-Cartesian which
    has same origin but axes are rotated by angle

    :param point: A point to be mapped
    :param angle: Rotation angle (in degrees) of x'y'-Cartesian coordinates
    :type point: pcbnew.wxPoint
    :type angle: float
    :return: Result position in x'y'-Cartesian coordinates
    :rtype: pcbnew.wxPoint
    """
    x, y = point.x, point.y
    angle = math.radians(angle)
    xr = (x * math.cos(angle)) + (y * math.sin(angle))
    yr = (-x * math.sin(angle)) + (y * math.cos(angle))
    return pcbnew.wxPoint(xr, yr)


def position_in_cartesian_coordinates(
    point: pcbnew.wxPoint, angle: float
) -> pcbnew.wxPoint:
    """Performs inverse operation to position_in_rotated_coordinates i.e.
    map position in rotated (by angle) x'y'-Cartesian to xy-Cartesian

    :param point: A point to be mapped
    :param angle: Rotation angle (in degrees) of x'y'-Cartesian coordinates
    :type point: pcbnew.wxPoint
    :type angle: float
    :return: Result position in xy-Cartesian coordinates
    :rtype: pcbnew.wxPoint
    """
    xr, yr = point.x, point.y
    angle = math.radians(angle)
    x = (xr * math.cos(angle)) - (yr * math.sin(angle))
    y = (xr * math.sin(angle)) + (yr * math.cos(angle))
    return pcbnew.wxPoint(x, y)


class KeyPlacer(BoardModifier):
    def __init__(
        self,
        logger: Logger,
        board: pcbnew.BOARD,
        key_distance: Tuple[float, float] = (19.05, 19.05),
    ) -> None:
        super().__init__(logger, board)

        self.__key_distance_x = pcbnew.FromMM(key_distance[0])
        self.__key_distance_y = pcbnew.FromMM(key_distance[1])
        self.logger.debug(
            f"Set key 1U distance: {self.__key_distance_x}/{self.__key_distance_y}"
        )
        self.__current_key = 1
        self.__reference_coordinate = pcbnew.wxPointMM(25, 25)

    def get_current_key(self, key_format: str) -> pcbnew.FOOTPRINT:
        return self.get_footprint(key_format.format(self.__current_key))

    def get_current_footprint(self, diode_format: str) -> Optional[pcbnew.FOOTPRINT]:
        try:
            f = self.get_footprint(diode_format.format(self.__current_key))
        except:
            f = None
        return f

    def calculate_corner_position_of_switch_diode_route(
        self, diode_pad_position: pcbnew.wxPoint, switch_pad_position: pcbnew.wxPoint
    ) -> pcbnew.wxPoint:
        x_diff = diode_pad_position.x - switch_pad_position.x
        y_diff = diode_pad_position.y - switch_pad_position.y
        if builtins.abs(x_diff) < builtins.abs(y_diff):
            up_or_down = -1 if y_diff > 0 else 1
            return pcbnew.wxPoint(
                diode_pad_position.x - x_diff,
                diode_pad_position.y + (up_or_down * builtins.abs(x_diff)),
            )
        else:
            left_or_right = -1 if x_diff > 0 else 1
            return pcbnew.wxPoint(
                diode_pad_position.x + (left_or_right * builtins.abs(y_diff)),
                diode_pad_position.y - y_diff,
            )

    def route_switch_with_diode(
        self,
        switch: pcbnew.FOOTPRINT,
        diode: pcbnew.FOOTPRINT,
        angle: float,
        template_track_points=None,
    ) -> None:
        """Performs routing between switch and diode elements.
        Assumes col-to-row configuration where diode anode is pad number '2'.

        :param switch: Switch footprint to be routed.
        :param diode: Diode footprint to be routed.
        :param angle: Rotation angle (in degrees) of switch footprint
                      (diode rotation is assumed to be the same)
        :param templateTrackPoints: List of positions (relative to diode pad position)
                                    of track corners connecting switch and diode.
                                    Does not support vias, will be routed on the layer
                                    of the diode.
                                    If None, use automatic routing algorithm.
        :type switch: FOOTPRINT
        :type diode: FOOTPRINT
        :type angle: float
        :type templateTrackPoints: List[pcbnew.wxPoint]
        """
        self.logger.info(f"Routing {switch.GetReference()} with {diode.GetReference()}")

        layer = pcbnew.B_Cu if self.get_side(diode) == Side.BACK else pcbnew.F_Cu
        switch_pad_position = switch.FindPadByNumber("2").GetPosition()
        diode_pad_position = diode.FindPadByNumber("2").GetPosition()
        if KICAD_VERSION >= (7, 0, 0):
            switch_pad_position = pcbnew.wxPoint(
                switch_pad_position.x, switch_pad_position.y
            )
            diode_pad_position = pcbnew.wxPoint(
                diode_pad_position.x, diode_pad_position.y
            )

        self.logger.debug(
            f"switchPadPosition: {switch_pad_position}, "
            f"diodePadPosition: {diode_pad_position}",
        )

        if template_track_points:
            if angle != 0:
                self.logger.info(f"Routing at {angle} degree angle")
            start = diode_pad_position
            for t in template_track_points:
                if angle != 0:
                    diode_pad_position_r = position_in_rotated_coordinates(
                        diode_pad_position, angle
                    )
                    end = t.__add__(diode_pad_position_r)
                    end = position_in_cartesian_coordinates(end, angle)
                else:
                    end = t.__add__(diode_pad_position)
                if end := self.add_track_segment_by_points(start, end, layer):
                    start = end
        elif (
            switch_pad_position.x == diode_pad_position.x
            or switch_pad_position.y == diode_pad_position.y
        ):
            self.add_track_segment_by_points(
                diode_pad_position, switch_pad_position, layer
            )
        else:
            # pads are not in single line, attempt routing with two segment track
            if angle != 0:
                self.logger.info(f"Routing at {angle} degree angle")
                switch_pad_position_r = position_in_rotated_coordinates(
                    switch_pad_position, angle
                )
                diode_pad_position_r = position_in_rotated_coordinates(
                    diode_pad_position, angle
                )

                self.logger.debug(
                    "In rotated coordinates: "
                    f"switchPadPosition: {switch_pad_position_r}, "
                    f"diodePadPosition: {diode_pad_position_r}",
                )

                corner = self.calculate_corner_position_of_switch_diode_route(
                    diode_pad_position_r, switch_pad_position_r
                )
                corner = position_in_cartesian_coordinates(corner, angle)
            else:
                corner = self.calculate_corner_position_of_switch_diode_route(
                    diode_pad_position, switch_pad_position
                )

            # first segment: at 45 degree angle
            # (might be in rotated coordinate system) towards switch pad
            self.add_track_segment_by_points(diode_pad_position, corner, layer)
            # second segment: up to switch pad
            self.add_track_segment_by_points(corner, switch_pad_position, layer)

    def get_current_relative_element_position(
        self, key_format: str, element_format: str
    ) -> ElementPosition:
        key1 = self.get_footprint(key_format.format(1))
        element1 = self.get_footprint(element_format.format(1))
        pos1 = self.get_position(key1)
        pos2 = self.get_position(element1)
        return ElementPosition(
            Point(pcbnew.ToMM(pos2.x - pos1.x), pcbnew.ToMM(pos2.y - pos1.y)),
            element1.GetOrientationDegrees(),
            self.get_side(element1),
        )

    def remove_dangling_tracks(self) -> None:
        connectivity = self.get_connectivity()

        def _is_dangling(track):
            if KICAD_VERSION >= (7, 0, 7):
                return connectivity.TestTrackEndpointDangling(track, False)
            return connectivity.TestTrackEndpointDangling(track)

        for track in self.board.GetTracks():
            if _is_dangling(track):
                self.board.RemoveNative(track)

    def check_if_diode_routed(
        self, key_format: str, diode_format: str
    ) -> list[pcbnew.wxPoint]:
        switch = self.get_footprint(key_format.format(1))
        diode = self.get_footprint(diode_format.format(1))
        net1 = switch.FindPadByNumber("2").GetNetname()
        net2 = diode.FindPadByNumber("2").GetNetname()
        tracks = [t for t in self.board.GetTracks() if t.GetNetname() == net1 == net2]

        # convert tracks to list of vectors which will be used
        # by `AddTrackSegmentByPoints`
        switch_pad_position = switch.FindPadByNumber("2").GetPosition()
        diode_pad_position = diode.FindPadByNumber("2").GetPosition()
        if KICAD_VERSION >= (7, 0, 0):
            switch_pad_position = pcbnew.wxPoint(
                switch_pad_position.x, switch_pad_position.y
            )
            diode_pad_position = pcbnew.wxPoint(
                diode_pad_position.x, diode_pad_position.y
            )

        points_sorted = []
        search_point = diode_pad_position
        for _ in range(0, len(tracks) + 1):
            for t in list(tracks):
                start = t.GetStart()
                end = t.GetEnd()
                if KICAD_VERSION >= (7, 0, 0):
                    start = pcbnew.wxPoint(start.x, start.y)
                    end = pcbnew.wxPoint(end.x, end.y)
                found_start = start.__eq__(search_point)
                found_end = end.__eq__(search_point)
                if found_start or found_end:
                    points_sorted.append(search_point)
                    search_point = end if found_start else start
                    tracks.remove(t)
                    self.board.RemoveNative(t)
                    break
        if points_sorted:
            points_sorted.pop(0)
            points_sorted.append(switch_pad_position)

        reduced_points = [p.__sub__(diode_pad_position) for p in points_sorted]
        self.logger.info(f"Detected template switch-to-diode path: {reduced_points}")
        return reduced_points

    def run(
        self,
        layout: dict,
        key_format: str,
        diode_info: Optional[ElementInfo],
        route_tracks: bool = False,
        additional_elements: List[ElementInfo] = [],
    ) -> None:
        diode_format = ""
        template_tracks = []
        column_switch_pads = {}
        row_diode_pads = {}

        self.logger.info(f"User layout: {layout}")
        keyboard = get_keyboard(layout)

        if diode_info:
            self.logger.info(f"Diode info: {diode_info}")
            diode_format = diode_info.annotation_format
            if route_tracks:
                # check if first switch-diode pair is already routed, if yes,
                # then reuse its track shape for remaining pairs,
                # otherwise try to use automatic 'router'
                template_tracks = self.check_if_diode_routed(key_format, diode_format)

        if diode_info:
            additional_elements = [diode_info] + additional_elements

        for element_info in additional_elements:
            if element_info.position_option == PositionOption.CURRENT_RELATIVE:
                position = self.get_current_relative_element_position(
                    key_format, element_info.annotation_format
                )
                element_info.position = position

        if type(self.__key_distance_x) != int or type(self.__key_distance_y) != int:
            # this should never happen, add check to satisfy type hints
            msg = "Unsupported key_distance type"
            raise RuntimeError(msg)

        for key in keyboard.keys:
            switch_footprint = self.get_current_key(key_format)

            width = key.width
            height = key.height
            position = (
                pcbnew.wxPoint(
                    (self.__key_distance_x * key.x)
                    + (self.__key_distance_x * width // 2),
                    (self.__key_distance_y * key.y)
                    + (self.__key_distance_y * height // 2),
                )
                + self.__reference_coordinate
            )
            self.set_position(switch_footprint, position)
            self.reset_rotation(switch_footprint)

            for element_info in additional_elements:
                annotation_format = element_info.annotation_format
                element_position = element_info.position
                footprint = self.get_current_footprint(annotation_format)
                if footprint and element_position:
                    self.reset_rotation(footprint)
                    self.set_side(footprint, element_position.side)
                    footprint.SetOrientationDegrees(element_position.orientation)
                    self.set_relative_position_mm(
                        footprint,
                        position,
                        element_position.relative_position.to_list(),
                    )

            angle = key.rotation_angle
            if angle != 0:
                rotation_reference = (
                    pcbnew.wxPoint(
                        (self.__key_distance_x * key.rotation_x),
                        (self.__key_distance_y * key.rotation_y),
                    )
                    + self.__reference_coordinate
                )
                self.rotate(switch_footprint, rotation_reference, angle)
                if additional_elements:
                    for element_info in additional_elements:
                        annotation_format = element_info.annotation_format
                        if footprint := self.get_current_footprint(annotation_format):
                            self.rotate(footprint, rotation_reference, angle)

            # append pad:
            pad = switch_footprint.FindPadByNumber("1")
            net_name = pad.GetNetname()
            if match := re.match(r"^COL(\d+)$", net_name):
                column_number = match.groups()[0]
                column_switch_pads.setdefault(column_number, []).append(pad)
            else:
                self.logger.warning("Switch pad without recognized net name found.")

            # append diode:
            diode_footprint = self.get_current_footprint(diode_format)
            if diode_footprint:
                pad = diode_footprint.FindPadByNumber("1")
                net_name = pad.GetNetname()
                if match := re.match(r"^ROW(\d+)$", net_name):
                    row_number = match.groups()[0]
                    row_diode_pads.setdefault(row_number, []).append(pad)
                else:
                    self.logger.warning("Switch pad without recognized net name found.")

            if route_tracks and diode_footprint:
                self.route_switch_with_diode(
                    switch_footprint, diode_footprint, angle, template_tracks
                )

            self.__current_key += 1

        if route_tracks:
            # very naive routing approach, will fail in some scenarios:
            for pads in column_switch_pads.values():
                positions = [pad.GetPosition() for pad in pads]
                for pos1, pos2 in zip(positions, positions[1:]):
                    # connect two pads together
                    if pos1.x == pos2.x:
                        self.add_track_segment_by_points(pos1, pos2, layer=pcbnew.F_Cu)
                    else:
                        # two segment track
                        y_diff = builtins.abs(pos1.y - pos2.y)
                        x_diff = builtins.abs(pos1.x - pos2.x)
                        vector = [0, (y_diff - x_diff)]
                        if vector[1] <= 0:
                            self.logger.warning(
                                "Switch pad to far to route 2 segment track "
                                "with 45 degree angles"
                            )
                        elif last_position := self.add_track_segment(
                            pos1, vector, layer=pcbnew.F_Cu
                        ):
                            self.add_track_segment_by_points(
                                last_position, pos2, layer=pcbnew.F_Cu
                            )

            for row in row_diode_pads:
                pads = row_diode_pads[row]
                positions = [pad.GetPosition() for pad in pads]
                # we can assume that all diodes are on the same side:
                layer = (
                    pcbnew.B_Cu
                    if self.get_side(pads[0].GetParent()) == Side.BACK
                    else pcbnew.F_Cu
                )
                for pos1, pos2 in zip(positions, positions[1:]):
                    if pos1.y == pos2.y:
                        self.add_track_segment_by_points(pos1, pos2, layer)
                    else:
                        self.logger.warning(
                            "Automatic diode routing supported only when "
                            "diodes aligned vertically"
                        )

            self.remove_dangling_tracks()



from __future__ import annotations

from logging import Logger

import pcbnew

from .element_position import Side

# remove pre-release and build numbers (if present) and split to major-minor-patch tuple
KICAD_VERSION = tuple(map(int, ((pcbnew.Version().split("+")[0]).split("-")[0]).split(".")))
DEFAULT_CLEARANCE_MM = 0.25


def set_position(footprint: pcbnew.FOOTPRINT, position: pcbnew.wxPoint) -> None:
    if KICAD_VERSION >= (7, 0, 0):
        footprint.SetPosition(pcbnew.VECTOR2I(position.x, position.y))
    else:
        footprint.SetPosition(position)


def get_position(footprint: pcbnew.FOOTPRINT) -> pcbnew.wxPoint:
    position = footprint.GetPosition()
    if KICAD_VERSION >= (7, 0, 0):
        return pcbnew.wxPoint(position.x, position.y)
    return position


class BoardModifier:
    def __init__(self, logger: Logger, board: pcbnew.BOARD) -> None:
        self.logger = logger
        self.board = board

    def get_connectivity(self):
        self.board.BuildConnectivity()
        return self.board.GetConnectivity()

    def get_footprint(self, reference: str) -> pcbnew.FOOTPRINT:
        self.logger.info(f"Searching for {reference} footprint")
        footprint = self.board.FindFootprintByReference(reference)
        if footprint is None:
            self.logger.error("Footprint not found")
            msg = f"Cannot find footprint {reference}"
            raise Exception(msg)
        return footprint

    def set_position(
        self, footprint: pcbnew.FOOTPRINT, position: pcbnew.wxPoint
    ) -> None:
        self.logger.info(
            f"Setting {footprint.GetReference()} footprint position: {position}"
        )
        set_position(footprint, position)

    def set_position_by_points(
        self, footprint: pcbnew.FOOTPRINT, x: int, y: int
    ) -> None:
        self.set_position(footprint, pcbnew.wxPoint(x, y))

    def get_position(self, footprint: pcbnew.FOOTPRINT) -> pcbnew.wxPoint:
        position = get_position(footprint)
        self.logger.info(
            f"Getting {footprint.GetReference()} footprint position: {position}"
        )
        return position

    def set_relative_position_mm(
        self,
        footprint: pcbnew.FOOTPRINT,
        reference_point: pcbnew.wxPoint,
        direction: list[float],
    ) -> None:
        position = pcbnew.wxPoint(
            reference_point.x + pcbnew.FromMM(direction[0]),
            reference_point.y + pcbnew.FromMM(direction[1]),
        )
        self.set_position(footprint, position)

    def test_track_collision(self, track: pcbnew.PCB_TRACK) -> bool:
        collide_list = []
        track_shape = track.GetEffectiveShape()
        track_start = track.GetStart()
        track_end = track.GetEnd()
        track_net_code = track.GetNetCode()
        # connectivity needs to be last,
        # otherwise it will update track net name before we want it to:
        connectivity = self.get_connectivity()
        for f in self.board.GetFootprints():
            reference = f.GetReference()
            hull = f.GetBoundingHull()
            if hit_test_result := hull.Collide(track_shape):
                for p in f.Pads():
                    pad_name = p.GetName()
                    pad_shape = p.GetEffectiveShape()

                    # track has non default netlist set so we can skip
                    # collision detection for pad of same netlist:
                    if track_net_code != 0 and track_net_code == p.GetNetCode():
                        self.logger.debug(
                            f"Track collision ignored, pad {reference}:{pad_name} "
                            f"on same netlist: {track.GetNetname()}/{p.GetNetname()}"
                        )
                        continue

                    # if track starts or ends in pad then assume that
                    # this collision is expected, with the exception of case
                    # where track already has netlist set and it is different
                    # than pad's netlist
                    if p.HitTest(track_start) or p.HitTest(track_end):
                        if (
                            track_net_code != 0
                            and track_net_code != p.GetNetCode()
                            and p.IsOnLayer(track.GetLayer())
                        ):
                            self.logger.debug(
                                f"Track collide with pad {reference}:{pad_name}"
                            )
                            collide_list.append(p)
                        else:
                            self.logger.debug(
                                "Track collision ignored, track starts or ends "
                                f"in pad {reference}:{pad_name}"
                            )
                    else:
                        hit_test_result = pad_shape.Collide(
                            track_shape, pcbnew.FromMM(DEFAULT_CLEARANCE_MM)
                        )
                        on_same_layer = p.IsOnLayer(track.GetLayer())
                        if hit_test_result and on_same_layer:
                            self.logger.debug(
                                f"Track collide with pad {reference}:{pad_name}"
                            )
                            collide_list.append(p)
        # track ids to clear at the end:
        tracks_to_clear = []
        for t in self.board.GetTracks():
            # check collision if not itself, on same layer and with different netlist
            # (unless 'trackNetCode' is default '0' netlist):
            if (
                t.m_Uuid.__ne__(track.m_Uuid)
                and t.IsOnLayer(track.GetLayer())
                and (track_net_code != t.GetNetCode() or track_net_code == 0)
            ):
                track_uuid = t.m_Uuid.AsString()
                if (
                    track_start == t.GetStart()
                    or track_start == t.GetEnd()
                    or track_end == t.GetStart()
                    or track_end == t.GetEnd()
                ):
                    self.logger.debug(
                        "Track collision ignored, track starts or ends "
                        f"at the end of {track_uuid} track"
                    )
                    # ignoring one track means that we can ignore
                    # all other connected to it:
                    tracks_to_clear += [
                        x.m_Uuid for x in connectivity.GetConnectedTracks(t)
                    ]
                    # check if connection to this track clears pad collision:
                    connected_pads_ids = [
                        x.m_Uuid for x in connectivity.GetConnectedPads(t)
                    ]
                    for collision in list(collide_list):
                        if collision.m_Uuid in connected_pads_ids:
                            self.logger.debug(
                                "Pad collision removed due to connection with track "
                                "which leads to that pad"
                            )
                            collide_list.remove(collision)
                elif hit_test_result := t.GetEffectiveShape().Collide(
                    track_shape, pcbnew.FromMM(DEFAULT_CLEARANCE_MM)
                ):
                    self.logger.debug(f"Track collide with another track: {track_uuid}")
                    collide_list.append(t)
        for collision in list(collide_list):
            if collision.m_Uuid in tracks_to_clear:
                collision_uuid = collision.m_Uuid.AsString()
                self.logger.debug(
                    f"Track collision with {collision_uuid} removed due to "
                    "connection with track which leads to it"
                )
                collide_list.remove(collision)
        return len(collide_list) != 0

    def add_track_to_board(self, track: pcbnew.PCB_TRACK):
        """Add track to the board if track passes collision check.
        If track has no set netlist, it would get netlist of a pad
        or other track, on which it started or ended.
        Collision with element of the same netlist will be ignored
        unless it is default '0' netlist.
        This exception about '0' netlist is important because it helps
        to detect collisions with holes.

        :param track: A track to be added to board
        :return: End position of added track or None if failed to add.
        """
        if not self.test_track_collision(track):
            layer_name = self.board.GetLayerName(track.GetLayer())
            start = track.GetStart()
            stop = track.GetEnd()
            self.logger.info(
                f"Adding track segment ({layer_name}): [{start}, {stop}]",
            )
            self.board.Add(track)
            return stop
        else:
            self.logger.warning("Could not add track segment due to detected collision")
            return None

    def add_track_segment_by_points(
        self, start: pcbnew.wxPoint, end: pcbnew.wxPoint, layer=pcbnew.B_Cu
    ):
        track = pcbnew.PCB_TRACK(self.board)
        track.SetWidth(pcbnew.FromMM(0.25))
        track.SetLayer(layer)
        if KICAD_VERSION >= (7, 0, 0):
            track.SetStart(pcbnew.VECTOR2I(start.x, start.y))
            track.SetEnd(pcbnew.VECTOR2I(end.x, end.y))
        else:
            track.SetStart(start)
            track.SetEnd(end)
        return self.add_track_to_board(track)

    def add_track_segment(
        self, start: pcbnew.wxPoint, vector: list[int], layer=pcbnew.B_Cu
    ):
        end = pcbnew.wxPoint(start.x + vector[0], start.y + vector[1])
        return self.add_track_segment_by_points(start, end, layer)

    def reset_rotation(self, footprint: pcbnew.FOOTPRINT):
        footprint.SetOrientationDegrees(0)

    def rotate(
        self,
        footprint: pcbnew.FOOTPRINT,
        rotation_reference: pcbnew.wxPoint,
        angle: float,
    ) -> None:
        self.logger.info(
            f"Rotating {footprint.GetReference()} footprint: "
            f"rotationReference: {rotation_reference}, rotationAngle: {angle}"
        )
        if KICAD_VERSION >= (7, 0, 0):
            footprint.Rotate(
                pcbnew.VECTOR2I(rotation_reference.x, rotation_reference.y),
                pcbnew.EDA_ANGLE(angle * -1, pcbnew.DEGREES_T),
            )
        else:
            footprint.Rotate(rotation_reference, angle * -10)

    def set_side(self, footprint: pcbnew.FOOTPRINT, side: Side) -> None:
        if side ^ self.get_side(footprint):
            footprint.Flip(footprint.GetPosition(), False)

    def get_side(self, footprint: pcbnew.FOOTPRINT) -> Side:
        return Side(footprint.IsFlipped())



import argparse
import json
import logging
from typing import List

import pcbnew

from .defaults import DEFAULT_DIODE_POSITION, ZERO_POSITION
from .element_position import ElementInfo, ElementPosition, Point, PositionOption, Side
from .key_placer import KeyPlacer
from .template_copier import TemplateCopier


class ElementInfoAction(argparse.Action):
    def __call__(self, parser, namespace, values: str, option_string=None) -> None:
        try:
            value: ElementInfo = self.parse(values, option_string)
        except ValueError as e:
            raise argparse.ArgumentTypeError(str(e))
        setattr(namespace, self.dest, value)

    def parse(self, values: str, option_string) -> ElementInfo:
        tokens: list[str] = values.split()
        err = ""
        if len(tokens) not in [2, 6]:
            err = f"{option_string} invalid format."
            raise ValueError(err)
        else:
            annotation = tokens[0]
            if annotation.count("{}") != 1:
                err = (
                    f"'{annotation}' invalid annotation specifier, "
                    "it must contain exactly one '{}' placeholder."
                )
                raise ValueError(err)

            option = PositionOption.get(tokens[1])
            position = None
            if len(tokens) == 2:
                if option not in [
                    PositionOption.CURRENT_RELATIVE,
                    PositionOption.DEFAULT,
                ]:
                    err = (
                        f"{option_string} position option needs to be equal "
                        "CURRENT_RELATIVE or DEFAULT if position details not provided"
                    )
                    raise ValueError(err)
            elif option != PositionOption.CUSTOM:
                err = (
                    f"{option_string} position option needs to be equal CUSTOM "
                    "when providing position details"
                )
                raise ValueError(err)
            else:
                floats = tuple(map(float, tokens[2:5]))
                side = Side.get(tokens[5])
                position = ElementPosition(Point(floats[0], floats[1]), floats[2], side)

            value: ElementInfo = ElementInfo(
                annotation,
                option,
                position,
            )
            return value


class ElementInfoListAction(ElementInfoAction):
    def __call__(self, parser, namespace, values: str, option_string=None) -> None:
        try:
            if "DEFAULT" in values:
                msg = f"{option_string} does not support DEFAULT position"
                raise ValueError(msg)
            value: List[ElementInfo] = []
            tokens: list[str] = values.split(";")
            for token in tokens:
                element_info: ElementInfo = super().parse(token.strip(), option_string)
                value.append(element_info)
        except ValueError as e:
            raise argparse.ArgumentTypeError(str(e))
        setattr(namespace, self.dest, value)


class XYAction(argparse.Action):
    def __call__(self, parser, namespace, values: str, option_string=None) -> None:
        try:
            value = tuple(map(float, values.split()))
            if len(value) != 2:
                msg = (
                    f"{option_string} must be exactly two numeric values "
                    "separated by a space."
                )
                raise ValueError(msg)
        except ValueError as e:
            raise argparse.ArgumentTypeError(str(e))
        setattr(namespace, self.dest, value)


def app():
    parser = argparse.ArgumentParser(
        description="Keyboard's key autoplacer",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    subparsers = parser.add_subparsers(dest="mode", help="Selects operation mode")
    parser_cli = subparsers.add_parser(
        "cli",
        help="Run in command-line mode",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser_gui = subparsers.add_parser(
        "gui",
        help="Run in graphical mode",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    for p in [parser_cli, parser_gui]:
        p.add_argument(
            "-b", "--board", required=True, help=".kicad_pcb file to be processed"
        )
    parser_cli.add_argument(
        "-l", "--layout", required=True, help="json layout definition file"
    )
    parser_cli.add_argument(
        "-r", "--route", action="store_true", help="Enable experimental routing"
    )
    parser_cli.add_argument(
        "-d",
        "--diode",
        default=ElementInfo("D{}", PositionOption.DEFAULT, DEFAULT_DIODE_POSITION),
        action=ElementInfoAction,
        help=(
            "Diode information, space separated value of ANNOTATION OPTION [POSITION]\n"
            "Available OPTION choices: DEFAULT, CURRENT_RELATIVE and CUSTOM\n"
            "When DEFAULT or CURRENT_RELATIVE, then POSITION needs to be omitted,\n"
            "when CUSTOM, then POSITION is space separated value of X Y ORIENTATION FRONT|BACK\n"
            "for example:\n"
            "\tD{} CURRENT_RELATIVE\n"
            "\tD{} CUSTOM 5 -4.5 90 BACK\n"
            "equal 'D{} DEFAULT' by default"
        ),
    )
    parser_cli.add_argument(
        "--additional-elements",
        default=[
            ElementInfo(
                "ST{}",
                PositionOption.CUSTOM,
                ZERO_POSITION,
            )
        ],
        action=ElementInfoListAction,
        help=(
            "List of ';' separated additional elements ELEMENT_INFO values\n"
            "ELEMENT_INFO is space separated value of ANNOTATION OPTION POSITION\n"
            "Available OPTION choices: CURRENT_RELATIVE and CUSTOM\n"
            "When CURRENT_RELATIVE, then POSITION needs to be omitted,\n"
            "when CUSTOM, then POSITION is space separated value of X Y ORIENTATION FRONT|BACK\n"
            "for example:\n"
            "\tST{} CUSTOM 0 0 180 BACK;LED{} CURRENT_RELATIVE\n"
            "equal 'ST{} CUSTOM 0 0 0 FRONT' by default"
        ),
    )
    parser_cli.add_argument(
        "--key-distance",
        default=(19.05, 19.05),
        action=XYAction,
        help=(
            "X and Y key 1U distance in mm, as two space separated numeric values, "
            "19.05 19.05 by default"
        ),
    )
    parser_cli.add_argument("-t", "--template", help="Controller circuit template")

    args = parser.parse_args()

    if not args.mode:
        parser.print_help()
        exit(0)
    elif args.mode == "cli":
        layout_path = args.layout
        board_path = args.board
        route_tracks = args.route
        diode = args.diode
        additional_elements = args.additional_elements
        key_distance = args.key_distance
        template_path = args.template
    else:
        # this must be gui mode (as other arguments would raise exception)
        import wx

        from .kbplacer_dialog import KbplacerDialog

        _ = wx.App()
        dlg = KbplacerDialog(None, "kbplacer")
        if dlg.ShowModal() == wx.ID_OK:
            layout_path = dlg.get_layout_path()
            board_path = args.board
            route_tracks = dlg.is_tracks()
            diode = dlg.get_diode_position_info()
            additional_elements = dlg.get_additional_elements_info()
            key_distance = dlg.get_key_distance()
            template_path = dlg.get_template_path()
        else:
            exit(0)

    # set up logger
    logging.basicConfig(
        level=logging.DEBUG, format="%(asctime)s: %(message)s", datefmt="%H:%M:%S"
    )
    logger = logging.getLogger(__name__)

    board = pcbnew.LoadBoard(board_path)

    if layout_path:
        with open(layout_path, "r") as footprint:
            text_input = footprint.read()
            layout = json.loads(text_input)

        placer = KeyPlacer(logger, board, key_distance)
        placer.run(
            layout,
            "SW{}",
            diode,
            route_tracks,
            additional_elements=additional_elements,
        )

    if template_path:
        copier = TemplateCopier(logger, board, template_path, route_tracks)
        copier.run()

    pcbnew.Refresh()
    pcbnew.SaveBoard(board_path, board)

    logging.shutdown()


if __name__ == "__main__":
    app()
